package mvd.jester;

import java.io.PrintWriter;
import java.lang.reflect.Constructor;
import java.time.LocalTime;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import mvd.jester.model.SystemSetup;
import mvd.jester.simulator.AbstractSimulator;
import mvd.jester.tests.AbstractTest;
import mvd.jester.utils.Logger;

/**
 * TestEnvironment
 */
public class TestEnvironment {

    private final long numberOfTaskSets;
    private final SystemSetup systemSetup;
    private final SystemSetup.Builder builder;
    private final Map<Constructor<? extends AbstractTest>, Constructor<? extends AbstractSimulator>> abstractTestPairs;

    public TestEnvironment(SystemSetup.Builder builder, long numberOfTaskSets) {
        this.numberOfTaskSets = numberOfTaskSets;
        abstractTestPairs = new HashMap<>();

        this.builder = builder;
        this.systemSetup = builder.build();
    }

    public TestEnvironment registerTestPair(Class<? extends AbstractTest> abstractTest,
            Class<? extends AbstractSimulator> abstractSimulator) {
        try {
            abstractTestPairs.put(abstractTest.getConstructor(SystemSetup.class),
                    abstractSimulator.getConstructor(SystemSetup.class));
        } catch (Exception e) {
            System.out.println("Missing constructor!");
        }

        return this;
    }

    public void runTests() {
        Map<AbstractTest, AbstractSimulator> testPairs = new HashMap<>();

        Map<Long, Map<AbstractTest, Long>> testResults = new HashMap<>();
        Map<Long, Map<AbstractSimulator, Long>> simulatorResults = new HashMap<>();

        for (Map.Entry<Constructor<? extends AbstractTest>, Constructor<? extends AbstractSimulator>> kv : abstractTestPairs
                .entrySet()) {
            try {
                testPairs.put(kv.getKey().newInstance(this.systemSetup),
                        kv.getValue().newInstance(this.systemSetup));
            } catch (Exception e) {
                System.out.println("Could not instantiate object of AbstractTest!");
            }
        }

        long checkedTasksets = 0;

        while (checkedTasksets < numberOfTaskSets) {
            builder.rebuild(this.systemSetup);

            double utilization = this.systemSetup.getUtilization();

            while (utilization <= this.systemSetup.getNumberOfProcessors()
                    && checkedTasksets < numberOfTaskSets) {
                checkedTasksets++;

                if (checkedTasksets % 100 == 0) {
                    System.out.println(checkedTasksets + " bereits geschafft");
                }

                long roundedUtilization = (long) (utilization * 10);
                for (Map.Entry<AbstractTest, AbstractSimulator> kv : testPairs.entrySet()) {
                    boolean schedCheck = kv.getKey().runSchedulabilityCheck();
                    boolean simCheck = kv.getValue().runSimulation();
                    if (schedCheck) {
                        testResults
                                .computeIfAbsent(roundedUtilization,
                                        k -> new HashMap<AbstractTest, Long>())
                                .compute(kv.getKey(), (k, v) -> (v == null) ? 1 : v + 1);
                    }
                    if (simCheck) {
                        simulatorResults
                                .computeIfAbsent(roundedUtilization,
                                        k -> new HashMap<AbstractSimulator, Long>())
                                .compute(kv.getValue(), (k, v) -> (v == null) ? 1 : v + 1);
                    }
                    if (schedCheck == true && simCheck == false) {
                        boolean sched = kv.getKey().runSchedulabilityCheck();
                        boolean sim = kv.getValue().runSimulation();

                        if (sched == true && sim == false) {
                            try (PrintWriter out = new PrintWriter(
                                    "results/manualCheck" + checkedTasksets + ".txt")) {
                                out.println(systemSetup);
                            } catch (Exception e) {
                                System.out.println("Ähm something went horribly wrong!");
                            }
                        }
                    }
                }
                builder.addTask(systemSetup);

                utilization = this.systemSetup.getUtilization();
            }
        }

        logResults(testPairs.keySet(), testResults);
        logSimulations(new HashSet<AbstractSimulator>(testPairs.values()), simulatorResults);
    }

    private void logResults(Set<AbstractTest> testCases,
            Map<Long, Map<AbstractTest, Long>> results) {
        LocalTime date = LocalTime.now();
        Logger log = new Logger("./results/results_" + systemSetup.getNumberOfProcessors() + "_"
                + date.getHour() + ":" + date.getMinute() + ".txt");
        String firstLine = new String("Utilization");

        for (AbstractTest t : testCases) {
            firstLine = firstLine + "\t" + t.getName();
        }

        log.log(firstLine);

        for (Long util : results.keySet()) {
            String line = String.valueOf((double) util / 10);
            Map<AbstractTest, Long> tests = results.get(util);

            for (AbstractTest t : testCases) {
                if (tests.containsKey(t)) {
                    line += "\t" + tests.get(t);
                } else {
                    line += "\t" + "0";
                }
            }
            log.log(line);
        }

        log.finalize();
    }

    private void logSimulations(Set<AbstractSimulator> testCases,
            Map<Long, Map<AbstractSimulator, Long>> results) {
        LocalTime date = LocalTime.now();
        Logger log = new Logger("./results/results_sim_" + systemSetup.getNumberOfProcessors() + "_"
                + date.getHour() + ":" + date.getMinute() + ".txt");
        String firstLine = new String("Utilization");

        for (AbstractSimulator t : testCases) {
            firstLine = firstLine + "\t" + t.getName();
        }

        log.log(firstLine);

        for (Long util : results.keySet()) {
            String line = String.valueOf((double) util / 10);
            Map<AbstractSimulator, Long> tests = results.get(util);

            for (AbstractSimulator t : testCases) {
                if (tests.containsKey(t)) {
                    line += "\t" + tests.get(t);
                } else {
                    line += "\t" + "0";
                }
            }
            log.log(line);
        }

        log.finalize();
    }
}
